// Copyright 2019 terrier989@gmail.com.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'package:built_value/built_value.dart';
import 'package:built_value/serializer.dart';
import 'package:datastore/datastore.dart';
import 'package:dint/data/core.dart';
import 'package:meta/meta.dart';

part 'entity.g.dart';

/// Superclass for all entities.
@BuiltValue(instantiable: false)
abstract class Entity {
  @BuiltValueField(wireName: '@meta')
  @nullable
  Meta get meta;

  Future<Document> insert({Collection collection});

  Future<void> delete();

  Future<Document> update();

  Map<String, Object> toDatastoreData();

  static Entity fromDatastoreSnapshot(Snapshot snapshot) {
    if (snapshot.data == null) {
      return null;
    }
    final data = Map<String, Object>.from(snapshot.data);

    // Add document for deserializer
    data['@document'] = snapshot.document;

    final type = snapshot.data['@type'];
    if (type == null) {
      throw StateError('Data doesn\'t have property "@type"');
    }
    final serializer = datastoreSerializers.serializerForWireName(type);
    if (serializer == null) {
      throw StateError('Serializer not found for type "$type"');
    }
    return datastoreSerializers.deserialize(
      data,
      specifiedType: FullType(serializer.types.first),
    );
  }
}

/// Superclass for all entity queries.
abstract class EntityQuery {
  @nullable
  MetaQuery get meta;
}

///  Default implementations of [Entity] methods.
@BuiltValue(instantiable: false)
abstract class EntityMixin implements Entity {
  @override
  Future<Document> insert({Collection collection}) async {
    final data = toDatastoreData();
    var document = meta?.document;
    if (document == null) {
      if (collection == null) {
        throw ArgumentError.notNull('collection');
      }
      document = collection.newDocument();
    }
    await document.insert(data: data);
    return document;
  }

  @override
  Future<Document> update() async {
    final data = toDatastoreData();
    var document = meta?.document;
    if (document == null) {
      throw StateError("Can' update an entity that is not inserted yet.");
    }
    await document.update(data: data);
    return document;
  }

  @override
  Future<void> delete() {
    return meta?.document?.delete();
  }

  @override
  Map<String, Object> toDatastoreData() {
    return datastoreSerializers.serialize(this);
  }
}

/// Default implementations of [EntityQuery] methods.
@BuiltValue(instantiable: false)
abstract class EntityQueryMixin implements EntityQuery {
  @nullable
  @override
  MetaQuery get meta;

  @mustCallSuper
  Filter toDatastoreFilter() {
    throw UnimplementedError();
  }
}
