// Copyright 2019 terrier989@gmail.com.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'package:datastore/adapters.dart';
import 'package:datastore/datastore.dart';
import 'package:dint/data/account.dart';
import 'package:dint/data/business.dart';
import 'package:dint/data/core.dart';
import 'package:test/test.dart';

void main() {
  group('Entity', () {
    Datastore datastore;
    Document document;
    Product product;
    Product productWithDocument;

    setUp(() {
      // Set datastore
      datastore = MemoryDatastore();
      document = datastore.collection('product').document('example');

      // Set entity
      product = Product.build((b) {
        b.meta = MetaBuilder()
          ..updatedAt = DateTime.fromMillisecondsSinceEpoch(0, isUtc: true)
          ..owner = (AccountBuilder()
            ..name = 'user name'
            ..phoneNumber = PhoneNumber('000')
            ..emailAddress = EmailAddress(
              'example@gmail.com',
            ));
        b.price = Money.usd(3.14);
        b.description = Text('abc');
      });

      productWithDocument = product.rebuild((b) {
        b.meta ??= MetaBuilder();
        b.meta.document = document;
      });
    });

    test('toDatastoreDocument()', () {
      expect(
        product.meta.owner.emailAddress,
        EmailAddress('example@gmail.com'),
      );
      expect(
        product.toDatastoreData(),
        {
          '@type': 'Product',
          '@meta': {
            'owner': {
              'name': 'user name',
              'emailAddress': 'example@gmail.com',
              'phoneNumber': '000',
            },
            'updatedAt': '1970-01-01T00:00:00.000Z',
          },
          'description': 'abc',
          'price': 'USD 3.14',
        },
      );
    });

    test('toDatastoreDocument() with document', () {
      expect(
        productWithDocument.meta.document,
        document,
      );
      final data = productWithDocument.toDatastoreData();
      expect(data, isNotEmpty);
      expect(data['@meta'], isNotNull);
      expect(data['price'], 'USD 3.14');

      final meta = data['@meta'] as Map<String, Object>;
      expect(meta['document'], isNotNull);
    });

    test('save()', () async {
      final document = await productWithDocument.insert();
      final snapshot = await document.get();
      expect(snapshot.document, document);

      // Read
      final actual = Entity.fromDatastoreSnapshot(snapshot) as Product;
      expect(actual.description, Text('abc'));
      expect(
        actual.meta.document,
        document,
      );
      expect(
        actual.meta.updatedAt,
        DateTime.fromMillisecondsSinceEpoch(0, isUtc: true),
      );
      expect(actual, productWithDocument);
    });
  });
}
